import re
from decimal import Decimal, InvalidOperation

from babel.core import get_global
from babel.numbers import format_currency
from django.utils.safestring import mark_safe

from django_prices.utils.locale import get_locale_data


def get_currency_fraction(currency):
    fractions = get_global("currency_fractions")
    try:
        fraction = fractions[currency]
    except KeyError:
        fraction = fractions["DEFAULT"]
    return fraction[0]


def format_price(value, currency, html=False):
    """
    Format decimal value as currency
    """
    try:
        value = Decimal(value)
    except (TypeError, InvalidOperation):
        return ""

    locale, locale_code = get_locale_data()
    pattern = locale.currency_formats.get("standard").pattern

    if html:
        pattern = re.sub("(\xa4+)", '<span class="currency">\\1</span>', pattern)

    result = format_currency(value, currency, format=pattern, locale=locale_code)
    return mark_safe(result)



from decimal import Decimal
from functools import total_ordering

from django.core import validators
from django.db.models import Field
from prices import Money, TaxedMoney

from django_prices import forms


@total_ordering
class NonDatabaseFieldBase:
    """Base class for all fields that are not stored in the database."""

    empty_values = list(validators.EMPTY_VALUES)

    # Field flags
    auto_created = False
    blank = True
    concrete = False
    editable = False
    unique = False

    is_relation = False
    remote_field = None
    generated = False

    many_to_many = None
    many_to_one = None
    one_to_many = None
    one_to_one = None

    def __init__(self):
        self.column = None
        self.primary_key = False

        self.creation_counter = Field.creation_counter
        Field.creation_counter += 1

    def __eq__(self, other):
        if isinstance(other, (Field, NonDatabaseFieldBase)):
            return self.creation_counter == other.creation_counter
        return NotImplemented

    def __lt__(self, other):
        if isinstance(other, (Field, NonDatabaseFieldBase)):
            return self.creation_counter < other.creation_counter
        return NotImplemented

    def __hash__(self):
        return hash(self.creation_counter)

    def contribute_to_class(self, cls, name, **kwargs):
        self.attname = self.name = name
        self.model = cls
        cls._meta.add_field(self, private=True)
        setattr(cls, name, self)

    def clean(self, value, model_instance):
        # Shortcircut clean() because Django calls it on all fields with
        # is_relation = False
        return value


class MoneyField(NonDatabaseFieldBase):
    description = (
        "A field that combines an amount of money and currency code into Money"
        "It allows to store prices with different currencies in one database."
    )

    def __init__(
        self,
        amount_field="price_amount",
        currency_field="price_currency",
        verbose_name=None,
        **kwargs,
    ):
        super().__init__()
        self.amount_field = amount_field
        self.currency_field = currency_field
        self.verbose_name = verbose_name

    def __str__(self):
        return f"MoneyField(amount_field={self.amount_field}, currency_field={self.currency_field})"  # noqa: E501

    def __get__(self, instance, cls=None):
        if instance is None:
            return self

        amount = getattr(instance, self.amount_field)
        currency = getattr(instance, self.currency_field)
        if amount is not None and currency is not None:
            return Money(amount, currency)
        return self.get_default()

    def __set__(self, instance, value):
        amount = None
        currency = None
        if value is not None:
            amount = value.amount
            currency = value.currency
        setattr(instance, self.amount_field, amount)
        setattr(instance, self.currency_field, currency)

    def formfield(self, **kwargs):
        available_currencies = []
        if hasattr(self, "model"):
            available_currencies = self.model._meta.get_field(
                self.currency_field
            ).choices
        return forms.MoneyField(available_currencies=available_currencies)

    def get_default(self):
        default_currency = None
        default_amount = Decimal(0)
        if hasattr(self, "model"):
            default_currency = self.model._meta.get_field(
                self.currency_field
            ).get_default()
            default_amount = self.model._meta.get_field(self.amount_field).get_default()

        if default_amount is None:
            return None
        return Money(default_amount, default_currency)


class TaxedMoneyField(NonDatabaseFieldBase):
    description = "A field that combines net and gross fields values into TaxedMoney."

    def __init__(
        self,
        net_amount_field="price_amount_net",
        gross_amount_field="price_amount_gross",
        currency="currency",
        verbose_name=None,
        **kwargs,
    ):
        super().__init__()
        self.net_amount_field = net_amount_field
        self.gross_amount_field = gross_amount_field
        self.currency = currency
        self.verbose_name = verbose_name

    def __str__(self):
        return f"TaxedMoneyField(net_amount_field={self.net_amount_field}, gross_amount_field={self.gross_amount_field}, currency={self.currency})"  # noqa: E501

    def __get__(self, instance, cls=None):
        if instance is None:
            return self
        net_amount = getattr(instance, self.net_amount_field)
        gross_amount = getattr(instance, self.gross_amount_field)
        currency = getattr(instance, self.currency)
        if net_amount is None or gross_amount is None:
            return None
        return TaxedMoney(Money(net_amount, currency), Money(gross_amount, currency))

    def __set__(self, instance, value):
        net_amount = None
        gross_amount = None
        currency = None
        if value is not None:
            net_amount = value.net.amount
            gross_amount = value.gross.amount
            currency = value.currency
        setattr(instance, self.net_amount_field, net_amount)
        setattr(instance, self.gross_amount_field, gross_amount)
        setattr(instance, self.currency, currency)



import itertools
from decimal import Decimal

from babel.numbers import get_currency_symbol
from django import forms
from django.core.validators import ValidationError
from prices import Money

from django_prices.utils.locale import get_locale_data
from django_prices.validators import (
    MaxMoneyValidator,
    MinMoneyValidator,
    MoneyPrecisionValidator,
)
from django_prices.widgets import FixedCurrencyMoneyInput, MoneyInput

__all__ = ("MoneyField", "MoneyInput")


def _get_symbol(currency_code: str) -> str:
    _, locale_code = get_locale_data()
    return get_currency_symbol(currency_code, locale_code)


def _get_currency_choices(currencies: list[str]) -> list[tuple[str, str]]:
    """Generate choices for SelectField.
    As a label we are presenting currency symbol."""
    currency_choices = [(code, _get_symbol(code)) for code in currencies]
    return currency_choices


class MoneyField(forms.MultiValueField):
    """A field for Money objects.

    When only one currency is provided, it renders as FixedCurrencyMoneyInput,
    when more you will get MoneyInput which allow you to choose both amount
    and currency.

    `available_currencies` argument is expected to be list of currency codes."""

    def __init__(
        self,
        available_currencies: list[str],
        widget=None,
        max_values: list | None = None,
        min_values: list | None = None,
        max_digits: int = None,
        decimal_places: int = None,
        validators=(),
        *args,
        **kwargs,
    ):
        decimal_field = forms.DecimalField(
            max_digits=max_digits, decimal_places=decimal_places
        )

        choices = _get_currency_choices(available_currencies)
        fields = (decimal_field, forms.ChoiceField(choices=choices))

        if widget is not None:
            raise NotImplementedError("Custom widgets are not supported by MoneyField.")

        if len(available_currencies) == 0:
            raise ValueError("At least one currency needed.")
        elif len(available_currencies) == 1:
            widget_instance = FixedCurrencyMoneyInput(currency=available_currencies[0])
        else:
            widget_instance = MoneyInput(
                choices=_get_currency_choices(available_currencies)
            )

        super().__init__(fields, widget=widget_instance, *args, **kwargs)

        self.validators = list(itertools.chain(self.default_validators, validators))
        self.validators.append(MoneyPrecisionValidator(max_digits, decimal_places))
        if max_values is not None:
            self.validators.extend(
                [MaxMoneyValidator(limit_value) for limit_value in max_values]
            )
        if min_values is not None:
            self.validators.extend(
                [MinMoneyValidator(limit_value) for limit_value in min_values]
            )

    def compress(self, data_list: tuple[Decimal, str] | None) -> Money | None:
        """If field is optional, return None when there is no amount provided."""
        if not data_list:
            return None

        amount, currency = data_list
        if amount in self.empty_values and self.required:
            raise ValidationError("Enter a valid amount of money")
        if currency in self.empty_values and self.required:
            raise ValidationError("Enter a valid currency code")

        if amount is None:
            return None

        return Money(amount, currency)



from babel.core import Locale, UnknownLocaleError
from django.conf import settings
from django.utils.translation import get_language, to_locale


def get_locale_data():
    language = get_language()
    if not language:
        language = settings.LANGUAGE_CODE
    locale_code = to_locale(language)
    locale = None
    try:
        locale = Locale.parse(locale_code)
    except (ValueError, UnknownLocaleError):
        # Invalid format or unknown locale
        # Fallback to the default language
        language = settings.LANGUAGE_CODE
        locale_code = to_locale(language)
        locale = Locale.parse(locale_code)
    return locale, locale_code



from babel import core as babel_core
from babel import support as babel_support
from django import template
from django.conf import settings
from django.utils.translation import get_language, to_locale

try:
    from pytz import timezone
except ImportError:
    timezone = None

from django_prices.utils.formatting import format_price

register = template.Library()


@register.filter
def amount(obj, format="text"):
    if format == "text":
        return format_price(obj.amount, obj.currency, html=False)
    if format == "html":
        return format_price(obj.amount, obj.currency, html=True)
    return _get_format().currency(obj.amount, obj.currency)


@register.filter
def discount_amount_for(discount, price):
    return discount(price) - price


def _get_format():
    locale = babel_core.Locale.parse(to_locale(get_language()))
    if timezone:
        tzinfo = timezone(settings.TIME_ZONE)
    else:
        tzinfo = None
    return babel_support.Format(locale, tzinfo)



from django import forms
from django.template.loader import render_to_string
from prices import Money

__all__ = ["MoneyInput", "FixedCurrencyMoneyInput"]


class MoneyInput(forms.MultiWidget):
    def __init__(self, choices, attrs=None):
        widgets = [
            forms.TextInput(attrs={"type": "number", "step": "any"}),
            forms.Select(choices=choices),
        ]
        super().__init__(widgets, attrs)

    def decompress(self, value):
        if value and isinstance(value, Money):
            return [value.amount, value.currency]
        return [None, None]


class FixedCurrencyMoneyInput(forms.MultiWidget):
    template = "prices/widget.html"

    def __init__(self, currency, attrs=None):
        self.currency = currency
        widgets = [
            forms.TextInput(attrs={"type": "number", "step": "any"}),
            forms.HiddenInput(),
        ]
        super().__init__(widgets, attrs)

    def decompress(self, value):
        if value and isinstance(value, Money):
            return [value.amount, self.currency]
        return [None, None]

    def render(self, name, value, attrs=None, renderer=None):
        widget = super().render(name, value, attrs=attrs, renderer=renderer)
        return render_to_string(
            self.template, {"widget": widget, "value": value, "currency": self.currency}
        )



import warnings

from django import template

from django_prices.templatetags.prices import amount as new_amount
from django_prices.utils.formatting import (
    format_price as new_format_price,
)
from django_prices.utils.formatting import (
    get_currency_fraction as new_get_currency_fraction,
)

register = template.Library()


def deprecation_warning():
    warnings.warn(
        "Module `prices_i18n` is deprecated and will be removed in future version. Use `prices` instead."  # noqa: E501
    )


def get_currency_fraction(currency):
    deprecation_warning()
    return new_get_currency_fraction(currency)


def format_price(value, currency, html=False):
    deprecation_warning()
    return new_format_price(value, currency, html)


@register.filter
def amount(obj, format="text"):
    deprecation_warning()
    return new_amount(obj, format)



from babel.numbers import get_currency_precision, is_currency
from django.core.validators import (
    DecimalValidator,
    MaxValueValidator,
    MinValueValidator,
    ValidationError,
)

from django_prices.utils.formatting import format_price


class MoneyPrecisionValidator(DecimalValidator):
    def __init__(self, *args):
        super().__init__(*args)

    def __call__(self, other):
        value = other.amount
        currency = other.currency
        super().__call__(value)

        if is_currency(currency):
            currency_precision = get_currency_precision(currency)
            exponent = value.as_tuple()[-1]
            if exponent >= 0:
                decimals = 0
            else:
                decimals = abs(exponent)
            if decimals > currency_precision:
                raise ValidationError(
                    self.messages["max_decimal_places"],
                    code="max_decimal_places",
                    params={"max": currency_precision},
                )


class MoneyValueValidator:
    def __call__(self, value):
        cleaned = self.clean(value)
        if cleaned.currency != self.limit_value.currency:
            return
        if self.compare(cleaned, self.limit_value):
            currency = self.limit_value.currency
            params = {
                "limit_value": format_price(self.limit_value.amount, currency),
                "show_value": format_price(cleaned.amount, currency),
                "value": format_price(value.amount, currency),
            }
            raise ValidationError(self.message, code=self.code, params=params)


class MaxMoneyValidator(MoneyValueValidator, MaxValueValidator):
    pass


class MinMoneyValidator(MoneyValueValidator, MinValueValidator):
    pass
